
/* Socket45 - Web browser plug-in container utility
 *
 * For more information, visit the web site:
 *   http://home.mieweb.com/jason/socket45/
 *
 * --------------------------------------------------------------------------
 *
 * Copyright (C) 1998 Jason Summers
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the author be held liable for any damages arising from the
 * use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, YOU MUST INCLUDE AN ACKNOWLEDGMENT IN THE PRODUCT 
 *    DOCUMENTATION.
 * 2. Altered source versions must be plainly marked as such, and must not be 
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 * 
 * --------------------------------------------------------------------------
 *
 * Set tabs to 4 spaces for best results.
 * This is C code, but has lots of C++ style comments. So sue me.
 */


#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
#include <commctrl.h>

#include "resource.h"

#include "npapi.h"
#include "npupp.h"

#include "socket45.h"

#define PRG         "Socket45"
#define PRGVERSION  "1.0.5"
#define USERAGENT   PRG "/" PRGVERSION
#define HOMEPAGE    "http://home.mieweb.com/jason/socket45/"
#define BORDER 4


#define MAX_LOG      200
#define MAX_URL      2200
#define MAX_PARAMS   2200

#ifndef MAX_PATH
#define MAX_PATH     260
#endif


// Store instance-related functions in a convenient structure. This is
// currently not necessary, but it will be helpful if this code is ever
// modified to support multiple simultaneous instances of a plugin.
typedef struct {
	HWND hWnd;
	HMODULE dll;
	NPPluginFuncs Funcs;

	FARPROC getentrypoints;
	FARPROC initialize;
	FARPROC shutdown;

	char dllname[MAX_PATH];
	char datafile[MAX_PATH];
	char mimetype[128];

	int autourl;
	char url[MAX_URL];

	int autoagent;
	char useragent[128];

	uint16 mode;   // NP_FULL or NP_EMBED

	int autosize;  //  dynamically size embedded plug-ins to window size?
	int hidden;    //  for hidden plugins (no window)
	int width,height;  // not used if fullpage or autosize is on
	int showurlreq;    // display warning dialog boxes on unimplemented calls

	int statusbar;
	int pfunctblsize, nfunctblsize;
	int versionmajor, versionminor;

	char params[MAX_PARAMS];


	int16 argc;
	char *argn[128];
	char *argv[128];

	int where_are_we;
	int iswindowed;
	int istransparent;

	// The window structure must be static, to accommodate stupid plugins
	// which store a pointer to it and access it at a later time.
	NPWindow window;

} plugindata_t;


// globals:

HINSTANCE hInst;
HWND hWndParent;
HWND hWndMain;

HWND hWndLog;
HWND hWndList;   // the listbox in the log window
HWND hWndStBar;
int sbheight;
POINT scrollpos;
int logging=0;       // is the log window open?

// 'pl' (and many of these other globals) is used directly by lots of functions. 
// It probably ought to be passed as a parameter instead. Oh well.
plugindata_t* pl;
NPP pinst;
int pluginstate;
NPNetscapeFuncs  NFuncs;

char current_dllname[MAX_PATH];
char datafile_dir[MAX_PATH];
char filetoopen[MAX_PATH];      // from command line


/*
The basic window hierarchy:

        ----------------------
       |                 Main |
       |   -----------------  |
	   |  |          Parent | |
	   |  |  -------------  | |
	   |  | |      Plugin | | |
	   |  | |             | | |
	   |  | |             | | |
	   |  | |             | | |
	   |  |  -------------  | |
	   |   -----------------  |
	   |                      |
	   | -------------------- |
	   ||        Status bar  ||
	   | -------------------- |
	    ----------------------

  Windowless plugins don't have a Plugin window and draw directly onto "Parent".
*/



LRESULT CALLBACK WndProcParent(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK WndProcMain(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK WndProcPlugin(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK WndProcLog(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK DlgProcSetup(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK DlgProcAbout(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);


// *********************************************************************
// *********  boring utility functions

char *GetErrMsg(NPError err)
{
	static char msg[80];

	switch(err) {
	case NPERR_NO_ERROR:                    strcpy(msg,"No error"); break;
	case NPERR_GENERIC_ERROR:               strcpy(msg,"Gereric error"); break;
	case NPERR_INVALID_INSTANCE_ERROR:      strcpy(msg,"Invalid instance"); break;
	case NPERR_INVALID_FUNCTABLE_ERROR:     strcpy(msg,"Invalid function table"); break;
	case NPERR_MODULE_LOAD_FAILED_ERROR:    strcpy(msg,"Module load failed"); break;
	case NPERR_OUT_OF_MEMORY_ERROR:         strcpy(msg,"Out of memory"); break;
	case NPERR_INVALID_PLUGIN_ERROR:        strcpy(msg,"Invalid plug-in"); break;
	case NPERR_INVALID_PLUGIN_DIR_ERROR:    strcpy(msg,"Invalid plug-in Dir"); break;
	case NPERR_INCOMPATIBLE_VERSION_ERROR:  strcpy(msg,"Incompatible version"); break;
	case NPERR_INVALID_PARAM:               strcpy(msg,"Invalid param"); break;
	case NPERR_INVALID_URL:                 strcpy(msg,"Invalid URL"); break;
	case NPERR_FILE_NOT_FOUND:              strcpy(msg,"File not found"); break;
	case NPERR_NO_DATA:                     strcpy(msg,"No data"); break;
	case NPERR_STREAM_NOT_SEEKABLE:         strcpy(msg,"Stream not seekable"); break;
	default: sprintf(msg,"Unknown error code (%d)",err);
	}
	return msg;
}


#define MSG_E 0 // error
#define MSG_S 1 // severe
#define MSG_W 2 // warning
#define MSG_I 3 // information

void DisplayMsg(int severity, char *fmt, ...)
{
	va_list ap;
	static char buf[3072];
	char *t;
	int flags;

	va_start(ap, fmt);
	vsprintf(buf,fmt, ap);
	va_end(ap);

	switch(severity) {
	case MSG_W: t="Warning"; flags=MB_ICONWARNING;      break;
	case MSG_S: t="Error";   flags=MB_ICONERROR;        break;
	case MSG_I: t="Notice";  flags=MB_ICONINFORMATION;  break;
	default:    t="Error";   flags=MB_ICONWARNING;      break;
	}
	MessageBox(hWndMain,buf,t,MB_OK|flags);
}


// Add a message the the event log (if it is open)
void LogEvent(char *fmt, ...)
{
	va_list ap;
	static char buf[3072];
	int count;

	if(!logging) return;

	va_start(ap, fmt);
	vsprintf(buf,fmt, ap);
	va_end(ap);

	// If we have hit the max number of messages, delete the first one.
	while(1) {
		count=SendMessage(hWndList,LB_GETCOUNT,0,0);
		if(count<MAX_LOG) break;
		SendMessage(hWndList,LB_DELETESTRING,0,0);
	}

	SendMessage(hWndList,LB_ADDSTRING,0,(LPARAM)buf);
	SendMessage(hWndList,LB_SETCURSEL,(WPARAM)count,0);
}

// Given a full path+filename, returns a pointer to the beginning
// of the filename
char *BaseName(char *fn)
{
	char p,i;

	p=0;
	for(i=0;fn[i];i++) {
		if(fn[i]=='\\') p=i+1;
	}

	return &fn[p];
}

// Set the main window caption.
void Caption(char *p, char *d)
{
	static char buf[1024];

	if(p) {
		if(d) {
			sprintf(buf,"%s - %s - %s",BaseName(d),BaseName(p),PRG);
			SetWindowText(hWndMain,buf);
		}
		else {
			sprintf(buf,"%s - %s",BaseName(p),PRG);
			SetWindowText(hWndMain,buf);
		}
	}
	else {
		SetWindowText(hWndMain,PRG);
	}
}

// Given a filename with a full path, CD to its directory.
void ChangeToDirectoryOf(char *fn)
{
	int p,i;
	char d[MAX_PATH];

	strcpy(d,fn);
	p=0;
	for(i=0;d[i];i++) {   // find last '\'
		if(d[i]=='\\') p=i;
	}
	if(p<1) return;

	d[p]='\0';
	SetCurrentDirectory(d);
}

// Parse the parameters given to the plug-in and put then in an arg list.
// This needs work.
void ParseParams(char *params,int16 *argc, char **argn,char **argv)
{
	char *ptr[128];
	char *s;
	int i;
	int c;   // == (*argc);
	char paramsp[MAX_PARAMS];  // "processed" parameters (strtok is destructive)

	strcpy(paramsp,params);

	c=0;

	// first break the string into parameters, but not by name/value
	s=strtok(paramsp," \t\r\n");
	while(s) {
		if(strlen(s)) {  // skip over whitespace
			ptr[c]=s;
			c++;
		}
		s=strtok(NULL," \t\r\n");
	}

//	argn[0]=ptr[0];
//	argv[0]=ptr[0];

	// now break into parameters
	for(i=0;i<c;i++) {
		argn[i]=strtok(ptr[i],"=");
		argv[i]=strtok(NULL,"=");
	}

	(*argc)=(int16)c;
}

int GetFileExtension(char *fn,char *ext)
{
	int p,i;

	p=0;
	for(i=0;fn[i];i++) {
		if(fn[i]=='.') p=i;
		if(fn[i]=='\\') p=0;
	}

	if(!p) return 0;
	if(strlen(&fn[p+1])>31) return 0;

	strcpy(ext,&fn[p+1]);
	return 1;
}

// This probably needs work (to handle network paths, etc.), 
// but it's really not important.
void File2URL(char *fn, char *furl)
{
	int i;
	sprintf(furl,"file:///%s",fn);
	for(i=8;furl[i];i++) {
		if(furl[i]=='\\') furl[i]='/';
	}
	if(furl[9]==':') furl[9]='|';
}

// does this look like an absolute url?
int AbsoluteURL(const char *url)
{
	if(!_strnicmp(url,"https://",8)) return 1;
	if(!_strnicmp(url,"http://",7)) return 1;
	if(!_strnicmp(url,"telnet:",7)) return 1;
	if(!_strnicmp(url,"mailto:",7)) return 1;
	if(!_strnicmp(url,"ftp://",6)) return 1;
	if(!_strnicmp(url,"file:",5)) return 1;
	if(!_strnicmp(url,"news:",5)) return 1;
	return 0;
}


// *********************************************************************
// *********  program config functions


int RegisterClasses(void)
{
	WNDCLASS  wc;
	DWORD lasterror;

	wc.style = 0;
	wc.lpfnWndProc = (WNDPROC)WndProcMain;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon = LoadIcon(hInst,"ICONMAIN");
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName =  "MAINMENU";
	wc.lpszClassName = "SOCKET45MAIN";
	if(!RegisterClass(&wc)) {
		lasterror=GetLastError();
	}

	// CS_OWNDC is needed for windowless plug-ins
	wc.style = CS_HREDRAW|CS_VREDRAW|CS_OWNDC;
	wc.lpfnWndProc = (WNDPROC)WndProcParent;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(LTGRAY_BRUSH);
	wc.lpszMenuName =  NULL;
	wc.lpszClassName = "SOCKET45PARENT";
	if(!RegisterClass(&wc)) {
		lasterror=GetLastError();
	}

	// Class for the plugin window (that is handed to the plugin)
	// Netscape plugin windows don't have the REDRAW styles, but the browser
	// seems to simulate them
	wc.style = CS_HREDRAW|CS_VREDRAW;
	wc.lpfnWndProc = (WNDPROC)WndProcPlugin;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon = NULL;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName =  NULL;
	wc.lpszClassName = "SOCKET45PLUGIN";
	if(!RegisterClass(&wc)) {
		lasterror=GetLastError();
	}

	// event log window
	wc.style = 0;
	wc.lpfnWndProc = (WNDPROC)WndProcLog;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon =  LoadIcon(hInst,"LOGICON");;
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName =  NULL;
	wc.lpszClassName = "SOCKET45LOG";
	if(!RegisterClass(&wc)) {
		lasterror=GetLastError();
	}
	return 1;
}

// Reset to default settings
void FactorySettings(plugindata_t *pl)
{
	strcpy(pl->datafile,"");
	strcpy(pl->dllname,"");
	strcpy(pl->mimetype,"");

	strcpy(pl->url,"");
	pl->autourl=1;

	strcpy(pl->useragent,USERAGENT);
	pl->autoagent=1;

	pl->mode=NP_FULL;

	pl->autosize=1;
	pl->hidden=0;
	pl->width=500;
	pl->height=400;
	pl->showurlreq=1;

	pl->where_are_we=0;
	pl->iswindowed=1;
	pl->istransparent=1;  // windowless plugins are transparent by default

	pl->statusbar=1;
	pl->versionmajor= NP_VERSION_MAJOR;
	pl->versionminor= NP_VERSION_MINOR;
	pl->pfunctblsize= sizeof(pl->Funcs);     // (60)   // 0 = default
	pl->nfunctblsize= sizeof(NFuncs);     // (88)   // 0 = default

	strcpy(pl->params,"");

}

// Calls the plugin's setwindow function.
// Logs and checks errors
// Uses globals pl->Window, pinst, etc.
void MySetWindow(void)
{
	NPError err;

	if(pl->hidden) return;

	LogEvent("SetWindow%s %d %d %d %d", 
		(pl->window.type==NPWindowTypeWindow)?"":"(*)",
		pl->window.x, pl->window.y, pl->window.width, pl->window.height);

	err=pl->Funcs.setwindow(pinst, &(pl->window) );
	if(err != NPERR_NO_ERROR) {
		DisplayMsg(MSG_S,"SetWindow failed: %s",GetErrMsg(err));
	}

}

// *********************************************************************
// *********  window sizing & scrollbar functions

// This section is not written very well.

// Scroll bars are a pain.
// Fortunately, we only need them for fixed-size embedded plug-ins.

void set_scrollbars(int setpos)
{
	SCROLLINFO si;
	RECT r;

	if(pluginstate!=ST_RUNNING) return;
	if(pl->mode!=NP_EMBED || pl->autosize || pl->hidden) {
		// no scroll bars
		ShowScrollBar(hWndParent,SB_BOTH,FALSE);
		return;
	}

	GetClientRect(hWndParent,&r);

	if(scrollpos.x<0) scrollpos.x=0;
	if(scrollpos.y<0) scrollpos.y=0;


	if(pl->width + 2*BORDER > r.right) {
		if(pl->width+2*BORDER-scrollpos.x < r.right) 
			 scrollpos.x= pl->width+2*BORDER -r.right;
	}
	else scrollpos.x=0;

	if(pl->height + 2*BORDER > r.bottom) {
		if(pl->height+2*BORDER-scrollpos.y < r.bottom) 
			 scrollpos.y= pl->height+2*BORDER -r.bottom;
	}
	else scrollpos.y=0;

	si.cbSize=sizeof(SCROLLINFO);
	si.fMask=SIF_ALL;
	si.nMin=0;
	si.nMax=pl->width + 2*BORDER -1;
	si.nPage=r.right;
	si.nPos=scrollpos.x;
	si.nTrackPos=0;
	SetScrollInfo(hWndParent,SB_HORZ,&si,TRUE);

	si.nMax=pl->height + 2*BORDER -1;
	si.nPage=r.bottom;
	si.nPos=scrollpos.y;
	SetScrollInfo(hWndParent,SB_VERT,&si,TRUE);


	if(setpos) {
		if(pl->iswindowed) {
			if(pl->hWnd) {
				SetWindowPos(pl->hWnd, NULL, BORDER-scrollpos.x, BORDER-scrollpos.y, 0,0, 
					SWP_NOZORDER|SWP_NOSIZE);
			}
		}
		else {    // windowless plugin
			pl->window.x=BORDER-scrollpos.x;
			pl->window.y=BORDER-scrollpos.y;
			MySetWindow();
			InvalidateRect(hWndParent,NULL,TRUE);
		}
	}

}

// caller must already set r
void get_position(RECT *r, int *l, int *t, int *w, int *h)
{
	if(pl->mode==NP_EMBED && pl->autosize) {  // embed, 100%
		*l= BORDER+r->left;
		*t= BORDER+r->top;
		*w= r->right-r->left-BORDER*2;
		*h= r->bottom-r->top-BORDER*2-pl->statusbar*sbheight;
	}
	else if(pl->mode==NP_EMBED) {   // embed, fixed size
		*l= BORDER+r->left - scrollpos.x;
		*t= BORDER+r->top - scrollpos.y;
		*w=pl->width;
		*h=pl->height;
	}
	else {      // full page
		*l= r->left;
		*t= r->top;
		*w= r->right-r->left;
		*h= r->bottom-r->top-pl->statusbar*sbheight;
	}
}

void size_window(void)
{
	RECT r;
	int p_l, p_t, p_w, p_h;

	if(pluginstate==ST_RUNNING) {

		GetClientRect(hWndMain,&r);
		SetWindowPos(hWndParent, NULL, r.left, r.top,
			r.right-r.left,r.bottom-r.top-(sbheight*pl->statusbar), SWP_NOZORDER);

		SetWindowPos(hWndStBar, NULL, r.left, r.bottom-sbheight,
			r.right-r.left, sbheight, SWP_NOZORDER);
		ShowWindow(hWndStBar,pl->statusbar?SW_SHOW:SW_HIDE);

		get_position(&r, &p_l, &p_t, &p_w, &p_h);

		//LogEvent("Resize %d %d %d %d",p_l, p_t, p_w, p_h);

		if(pl->hWnd)
			SetWindowPos(pl->hWnd, NULL, p_l, p_t, p_w, p_h, SWP_NOZORDER);

		set_scrollbars(1);
	}
}

void send_setwindow(void)
{
	RECT r;
	int p_l, p_t, p_w, p_h;

	if(pluginstate!=ST_RUNNING) {
		MessageBox(hWndMain,"Internal: can't setwindow here","Error",MB_OK|MB_ICONERROR);
		return;
	}

	if(pl->hidden) return;

	GetClientRect(hWndMain,&r);

	get_position(&r, &p_l, &p_t, &p_w, &p_h);

	pl->window.x= p_l;
	pl->window.y= p_t;
	pl->window.width= p_w;
	pl->window.height= p_h;

	pl->window.clipRect.left=(uint16)0;
	pl->window.clipRect.top=(uint16)0;
	pl->window.clipRect.right=(uint16)0;
	pl->window.clipRect.bottom=(uint16)0;
	//pl->window.clipRect.left=(uint16)(p_l);   // shouldn't be necessary in Windows
	//pl->window.clipRect.top=(uint16)(p_t);
	//pl->window.clipRect.right=(uint16)(p_w);
	//pl->window.clipRect.bottom=(uint16)(p_h);

	if(pl->iswindowed) {
		pl->window.type=NPWindowTypeWindow;
		pl->window.window=(void*)pl->hWnd;
	}
	else {
		pl->window.type=NPWindowTypeDrawable;
		pl->window.window=(void*)GetDC(hWndParent);
	}

	MySetWindow();

}

// *********************************************************************
// *********  "the main cycle" of functions:
//  LoadPlugin
//  NewPluginInstance
//  WriteStream
//  DestroyPluginInstance
//  UnloadPlugin


// IE uses 30720. Netscape uses 10240 (larger for fullpage plugins?)
#define BUFFERSIZE 30720

// Load the .dll file and call Initialize
void LoadPlugin(void)
{
	NPError err;
	char buf[256],buf2[32];
	int i,c;
	int tblsize;
	void **v;

	if(pluginstate != ST_OFF) return;

	Caption(NULL,NULL);

	if(!strlen(pl->dllname)) {
		DisplayMsg(MSG_E,"No plug-in selected");
		return;
	}

	ChangeToDirectoryOf(pl->dllname);

	LogEvent("Loading plug-in: %s",pl->dllname);

	pl->dll=LoadLibrary(pl->dllname);
	if(!pl->dll) {
		DisplayMsg(MSG_S,"Can't load plug-in %s",pl->dllname);
		return;
	}

	pl->getentrypoints = GetProcAddress(pl->dll,(LPCSTR)1);
	pl->initialize     = GetProcAddress(pl->dll,(LPCSTR)2);
	pl->shutdown       = GetProcAddress(pl->dll,(LPCSTR)3);

	if(!pl->getentrypoints || !pl->initialize || !pl->shutdown) {
		DisplayMsg(MSG_S,"GetProcAddress failed");
		FreeLibrary(pl->dll);
		return;
	}

	// Call getentrypoints to retrieve pointers to the plugin's (NPP_*) functions.

	ZeroMemory(&pl->Funcs,sizeof(pl->Funcs));

	// Can emulate older browsers by providing space for fewer plugin functions.
	// IE4: 52 (last is javaclass)
	tblsize=(pl->pfunctblsize)?(pl->pfunctblsize):sizeof(pl->Funcs);
	pl->Funcs.size=tblsize;

	err=pl->getentrypoints(&pl->Funcs);
	if(err!=NPERR_NO_ERROR) {
		DisplayMsg(MSG_S,"GetEntryPoints failed: %s",GetErrMsg(err));
		FreeLibrary(pl->dll);
		return;
	}

	Caption(pl->dllname,NULL);

	LogEvent("Plug-in's API version: %d.%d",(int)(pl->Funcs.version>>8),(int)(pl->Funcs.version%0xff));

	// count the number of functions pointers the plugin actually filled in
	strcpy(buf,"");

	v=(void**)(&(pl->Funcs.newp));
	c=0;
	for(i=0;i<14;i++) {
		if( v[i] != NULL ) {
			sprintf(buf2,"%s%d",c?" ":"",i+1);
			strcat(buf,buf2);
			c++;
		}
	}
	LogEvent("Plug-in implements %d of %d functions (%s)",c,(tblsize-4)/4,buf);


	strcpy(current_dllname,pl->dllname);  // Save this (for logging purposes) as pl->dll may 
	                                      // change before we can unload it

	// Call initialize() to give the plugin pointers to the browser functions (NPN_*) that
	// it can call.

	NFuncs.size= (pl->nfunctblsize)?(pl->nfunctblsize):sizeof(NFuncs);
    NFuncs.version= (pl->versionmajor << 8) | pl->versionminor;
    NFuncs.geturl=            S45_GetURL;
    NFuncs.posturl=           S45_PostURL;
    NFuncs.requestread=       S45_RequestRead;
    NFuncs.newstream=         S45_NewStream;
    NFuncs.write=             S45_Write;
    NFuncs.destroystream=     S45_DestroyStream;
    NFuncs.status=            S45_Status;
    NFuncs.uagent=            S45_UserAgent;
    NFuncs.memalloc=          S45_MemAlloc;
    NFuncs.memfree=           S45_MemFree;
    NFuncs.memflush=          S45_MemFlush;
    NFuncs.reloadplugins=     S45_ReloadPlugins;
    NFuncs.getJavaEnv=        S45_GetJavaEnv;
    NFuncs.getJavaPeer=       S45_GetJavaPeer;
    NFuncs.geturlnotify=      S45_GetURLNotify;
    NFuncs.posturlnotify=     S45_PostURLNotify;
	NFuncs.getvalue=          S45_GetValue;
    NFuncs.setvalue=          S45_SetValue;
    NFuncs.invalidaterect=    S45_InvalidateRect;
    NFuncs.invalidateregion=  S45_InvalidateRegion;
    NFuncs.forceredraw=       S45_ForceRedraw;

	err = pl->initialize(&NFuncs);

	pluginstate=ST_LOADED;
}

// roughly corresponds to NPP_New
void NewPluginInstance(void)
{
	NPError err;
	static char *nullstr = "";

	if(pluginstate==ST_OFF) LoadPlugin();
	if(pluginstate!=ST_LOADED) return;

	LogEvent("Starting new plug-in instance");

	hWndParent = CreateWindow("SOCKET45PARENT",PRG,
		WS_CHILD|WS_VISIBLE|WS_OVERLAPPED|WS_HSCROLL|WS_VSCROLL|WS_CLIPCHILDREN,
		0,0,100,100,hWndMain,NULL,hInst,NULL);
	if (!hWndParent) return;

	/////////// call NPP_New

	pl->argc=0;
	pl->argn[0]=nullstr;
	pl->argv[0]=nullstr;

	if(pl->mode==NP_EMBED) {
		ParseParams(pl->params,&(pl->argc),pl->argn,pl->argv);
	}

	pinst->pdata=NULL;   // This will be changed by the plugin.
	pinst->ndata=pl;     // Not really needed since we'll only have 1 instance.

	pl->where_are_we=1;     // A little hack to make sure windowless plugins
	                        // declare themselves at the right time.
	err=pl->Funcs.newp(pl->mimetype, pinst, pl->mode, pl->argc, pl->argn, pl->argv, NULL);
	pl->where_are_we=0;

	if(err!=NPERR_NO_ERROR) {
		DisplayMsg(MSG_S,"NPP_New failed: %s",GetErrMsg(err));
		return;
	}

	////////////////////////
	if(pl->iswindowed && !pl->hidden) {

		pl->hWnd = CreateWindow(
			"SOCKET45PLUGIN","a Plugin Window",
			WS_CHILD|WS_OVERLAPPED,
			0,0,50,50,hWndParent,
			NULL,hInst,NULL);

		if(!pl->hWnd) {
			DisplayMsg(MSG_S,"Can't create window for plug-in");
			return;
		}
	}
	else {
		pl->hWnd = NULL;
	}

	pluginstate=ST_RUNNING;

	scrollpos.x=0; scrollpos.y=0;

	size_window();

	if(pl->hWnd) {
		ShowWindow(pl->hWnd, SW_SHOW); // now that everything's the right size, make it VISIBLE
	}

	send_setwindow();
}

// send the data file to the plug-in
void WriteStream(void)
{
	NPStream stream;
	NPError err;
	DWORD bytesleft, bytestotal;
	DWORD bytesread;
	HANDLE hFile;
	uint16 stype;
	int32 byteswanted, bytesaccepted;
	unsigned char *fbuf;
	char fileurl[MAX_PATH+16];
	int timeout;
	char buf[80];

	if(pluginstate!=ST_RUNNING) 
		NewPluginInstance();

	if(pluginstate!=ST_RUNNING)
		return;

	if(!strlen(pl->datafile)) return;

	// stream the data to the plug-in
	hFile=CreateFile(pl->datafile,GENERIC_READ,FILE_SHARE_READ,NULL,
		OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

	if(hFile==INVALID_HANDLE_VALUE) {
		DisplayMsg(MSG_S,"Can't open data file: %s",pl->datafile);
		return;
	}

	bytestotal=GetFileSize(hFile,NULL);
	bytesleft=bytestotal;

	stream.pdata=pinst->pdata;  // this seems a little superfluous (but some plugins rely on it)
	stream.ndata=pinst->ndata;
	if(pl->autourl) {
		File2URL(pl->datafile,fileurl);
		stream.url=fileurl;
	}
	else {
		stream.url=pl->url;
	}
	stream.end=bytestotal;
	stream.lastmodified=0;    // ??
	stream.notifyData=NULL;
	stype=NP_NORMAL;

	Caption(current_dllname,pl->datafile);

	err=pl->Funcs.newstream(pinst, pl->mimetype, &stream, 0, &stype);
	switch(stype) {
	case NP_NORMAL:      sprintf(buf,"NORMAL"); break;
	case NP_SEEK:        sprintf(buf,"SEEK (invalid)"); break;
	case NP_ASFILE:      sprintf(buf,"ASFILE"); break;
	case NP_ASFILEONLY:  sprintf(buf,"ASFILEONLY"); break;
	default:             sprintf(buf,"Unknown (%d)",(int)stype);
	}
	LogEvent("Streaming method: %s",buf);

	// Both NS and MSIE call setwindow again here for no apparent reason
	//send_setwindow();
	if(stype!=NP_ASFILEONLY) {
		LogEvent("Streaming file \"%s\" to plug-in",pl->datafile);

		fbuf=malloc(BUFFERSIZE);
		if(!fbuf) {
			DisplayMsg(MSG_S,"Out of memory");
			CloseHandle(hFile);
			return;
		}

		// This part should probably be asynchronous....
		timeout=0;

		while(bytesleft>0) {
			//Sleep(200); /////////////  might be nice to have a feature to slow it down...

			byteswanted=pl->Funcs.writeready(pinst, &stream);
			if(byteswanted<1) {
				Sleep(200);
				timeout++;
				if(timeout>15) {   // timeout after about 3 seconds
					DisplayMsg(MSG_W,"Plug-in is refusing to accept data; giving up");
					free(fbuf);
					CloseHandle(hFile);
					return;
				}

				continue;
			}
			timeout=0;

			ReadFile(hFile,fbuf,min(byteswanted,BUFFERSIZE),&bytesread,NULL);
			bytesaccepted=pl->Funcs.write(pinst, &stream, bytestotal-bytesleft, bytesread, fbuf);

			// If plugin returns a negative number, it doesn't want any more data.
			if(bytesaccepted<0) {
				LogEvent("Plug-in reported an error in processing stream data");
				break;
			}

			/* Netscape sez: 
			 *    Note that a plug-in must consume at least as many bytes as it indicated in the
			 *    preceeding NPP_WriteReady call. */
			if(bytesaccepted<(int32)bytesread) {
				DisplayMsg(MSG_E,"Plug-in did not accept enough data\r\n(Requested %d, accepted %d)",
					byteswanted, bytesaccepted);
			}
			bytesleft-=bytesread;
		}

		// a final WriteReady for no apparent reason (that's what NS and IE do)
		byteswanted=pl->Funcs.writeready(pinst, &stream);

		free(fbuf);
	}

	CloseHandle(hFile);

	// NPP_StreamAsFile if requested by the plugin
	if(stype==NP_ASFILE || stype==NP_ASFILEONLY) {
		LogEvent("Sending filename \"%s\" to plug-in",pl->datafile);
		pl->Funcs.asfile(pinst, &stream, pl->datafile);
	}

	// IE calls SetWindow again here for no reason
	//send_setwindow();

	// NPP_DestroyStream
	err=pl->Funcs.destroystream(pinst, &stream, NPRES_DONE);
	if(err!=NPERR_NO_ERROR) {
		DisplayMsg(MSG_S,"DestroyStream failed: %s",GetErrMsg(err));
	}
}


void DestroyPluginInstance(void)
{
	NPSavedData *sd;

	if(pluginstate != ST_RUNNING) return;

	pluginstate= ST_LOADED;

	LogEvent("Destroying plug-in instance");

	sd=NULL;
	pl->Funcs.destroy(pinst, &sd);

	if(sd) {  //        sd->len sd->buf
		LogEvent("Plug-in saved state data, length %d (at %p, %p)",sd->len,sd->buf,sd);
		// we don't support saved states (yet?), so free the memory the plug-in alloc'd for it
		if(sd->buf) free(sd->buf);
		free(sd);
	}

	if(pl->hWnd && IsWindow(pl->hWnd))
		DestroyWindow(pl->hWnd);

	ShowWindow(hWndStBar,SW_HIDE);
	SetWindowText(hWndStBar,"");

	if(IsWindow(hWndParent))
		DestroyWindow(hWndParent);

	Caption(current_dllname,NULL);
}

void UnloadPlugin(void)
{
	if(pluginstate==ST_OFF) return;

	if(pluginstate==ST_RUNNING) DestroyPluginInstance();

	if(pl->shutdown) {
		pl->shutdown();
	}

	if(pl->dll) {
		LogEvent("Unloading plug-in: %s",current_dllname);
		FreeLibrary(pl->dll);
		strcpy(current_dllname,"");
	}

	pluginstate=ST_OFF;
	Caption(NULL,NULL);
}


// *********************************************************************
// *********  user interface, etc.


void EraseSettings(HWND hWnd)
{
	char fn[MAX_PATH];
	char buf[256];
	LONG rv;

	strcpy(fn,BaseName(pl->datafile));
	if(!strlen(fn)) {
		DisplayMsg(MSG_E,"No valid file name chosen");
		return;
	}
	sprintf(buf,"SOFTWARE\\Generic\\Socket45\\files\\%s",fn);
	rv=RegDeleteKey(HKEY_CURRENT_USER,buf);
	if(rv==ERROR_SUCCESS) {
		DisplayMsg(MSG_I,"Erased settings for file '%s'",fn);
	}
	else {
		DisplayMsg(MSG_I,"Could not erase settings for file '%s'",fn);
	}
}

void SaveSettings(HWND hWnd, int bytype)
{
	DWORD disp;
	HKEY key;
	LONG rv;
	char szKey[256];
	char ext[256];
	DWORD tmpd;

	if(bytype) {
		if(!GetFileExtension(pl->datafile,ext)) {
			DisplayMsg(MSG_E,"Data file does not have a type (extension)");
			return;
		}
		sprintf(szKey,"SOFTWARE\\Generic\\Socket45\\types\\%s",ext);
	}
	else {
		strcpy(ext,BaseName(pl->datafile));

		if(!strlen(ext)) {
			DisplayMsg(MSG_E,"No valid file name to save");
			return;
		}
		sprintf(szKey,"SOFTWARE\\Generic\\Socket45\\files\\%s",ext);
	}

	rv=RegCreateKeyEx(
		HKEY_CURRENT_USER,szKey,
		(DWORD)0,(LPSTR)NULL,(DWORD)REG_OPTION_NON_VOLATILE,(REGSAM)KEY_ALL_ACCESS,
		NULL,&key,&disp);
	if(rv!=ERROR_SUCCESS) {
		DisplayMsg(MSG_E,"Unable to save settings!");
		return;
	}

	rv=RegSetValueEx(key,"plugin",0,REG_SZ,(LPBYTE)pl->dllname,1+strlen(pl->dllname));
	rv=RegSetValueEx(key,"mimetype",0,REG_SZ,(LPBYTE)pl->mimetype,1+strlen(pl->mimetype));

	rv=RegSetValueEx(key,"autourl",0,REG_DWORD,(LPBYTE)&(pl->autourl),sizeof(DWORD));
	rv=RegSetValueEx(key,"url",0,REG_SZ,(LPBYTE)pl->url,1+strlen(pl->url));

	rv=RegSetValueEx(key,"autoagent",0,REG_DWORD,(LPBYTE)&(pl->autoagent),sizeof(DWORD));
	rv=RegSetValueEx(key,"agent",0,REG_SZ,(LPBYTE)pl->useragent,1+strlen(pl->useragent));

	tmpd=(DWORD)pl->mode;
	rv=RegSetValueEx(key,"mode",0,REG_DWORD,(LPBYTE)&(tmpd),sizeof(DWORD));

	rv=RegSetValueEx(key,"showurlreq",0,REG_DWORD,(LPBYTE)&(pl->showurlreq),sizeof(DWORD));
	rv=RegSetValueEx(key,"autosize",0,REG_DWORD,(LPBYTE)&(pl->autosize),sizeof(DWORD));
	rv=RegSetValueEx(key,"hidden",0,REG_DWORD,(LPBYTE)&(pl->hidden),sizeof(DWORD));
	rv=RegSetValueEx(key,"width",0,REG_DWORD,(LPBYTE)&(pl->width),sizeof(DWORD));
	rv=RegSetValueEx(key,"height",0,REG_DWORD,(LPBYTE)&(pl->height),sizeof(DWORD));
	rv=RegSetValueEx(key,"statusbar",0,REG_DWORD,(LPBYTE)&(pl->statusbar),sizeof(DWORD));
	rv=RegSetValueEx(key,"versionmajor",0,REG_DWORD,(LPBYTE)&(pl->versionmajor),sizeof(DWORD));
	rv=RegSetValueEx(key,"versionminor",0,REG_DWORD,(LPBYTE)&(pl->versionminor),sizeof(DWORD));
	rv=RegSetValueEx(key,"pfunctblsize",0,REG_DWORD,(LPBYTE)&(pl->pfunctblsize),sizeof(DWORD));
	rv=RegSetValueEx(key,"nfunctblsize",0,REG_DWORD,(LPBYTE)&(pl->nfunctblsize),sizeof(DWORD));
	rv=RegSetValueEx(key,"params",0,REG_SZ,(LPBYTE)pl->params,1+strlen(pl->params));
	
	RegCloseKey(key);

	if(rv!=ERROR_SUCCESS) {
		DisplayMsg(MSG_E,"Unable to save settings!");
		return;
	}

	if(bytype)
		DisplayMsg(MSG_I,"Saved settings for file '%s'",ext);
	else
		DisplayMsg(MSG_I,"Saved settings for type '%s'",ext);

}

// Returns name in fn, but caller must initialize fn before calling
int GetDatafileName(HWND hWnd, char *fn)
{
	OPENFILENAME ofn;

	ZeroMemory(&ofn,sizeof(OPENFILENAME));

	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=hWnd;
	ofn.lpstrFilter="All files\0*.*\0\0";
	ofn.nFilterIndex=1;
	ofn.lpstrTitle="Select data file";
	ofn.lpstrFile=fn;
	if(strlen(datafile_dir)) ofn.lpstrInitialDir=datafile_dir;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

	if(GetOpenFileName(&ofn)) {
		strcpy(datafile_dir,fn);
		datafile_dir[ofn.nFileOffset]='\0';  // chop off the filename, leaving the dir
		return 1;
	}
	return 0;
}

int GetPluginName(HWND hWnd, char *fn)
{
	OPENFILENAME ofn;

	ZeroMemory(&ofn,sizeof(OPENFILENAME));

	ofn.lStructSize=sizeof(OPENFILENAME);
	ofn.hwndOwner=hWnd;
	ofn.lpstrFilter="Plug-ins (np*.dll)\0np*.dll\0All files\0*.*\0\0";
	ofn.nFilterIndex=1;
	ofn.lpstrTitle="Select Plug-in";
	ofn.lpstrFile=fn;
	ofn.nMaxFile=MAX_PATH;
	ofn.Flags=OFN_FILEMUSTEXIST|OFN_HIDEREADONLY;

	if(GetOpenFileName(&ofn)) {
		return 1;
	}
	return 0;
}

void PrintPlugin(HWND hWnd)
{
	PRINTDLG pd;
	DOCINFO di;
	HCURSOR hCur;
	NPPrint nppr;
	NPPrint1 nppr1;
	NPPrint2 nppr2;
	int done;

	if(pluginstate != ST_RUNNING) return;
	done=0;
	hCur=SetCursor(LoadCursor(NULL,IDC_WAIT));

	// If full-page, first give plugin the opportunity to 
	// "take over printing completely"
	if(pl->mode==NP_FULL) {
		ZeroMemory(&nppr,sizeof(NPPrint));
		nppr.mode=NP_FULL;
		nppr.print.fullPrint.pluginPrinted=FALSE;

		// TRUE if print button was pressed; FALSE if print menu (???)
		// (I don't know what the point of this is.)
		nppr.print.fullPrint.printOne=FALSE;

		// I don't know what this should be...
		nppr.print.fullPrint.platformPrint=NULL;
		pl->Funcs.print(pinst,&nppr);

		if(nppr.print.fullPrint.pluginPrinted) {
			// The plugin did everything. This pretty much never happens.
			// Probably because NS and IE don't support it correctly.
			done=1;
			LogEvent("Full-page plug-in printed itself");
		}
		else {
			LogEvent("Full-page plug-in declined to print itself");
		}
	}

	if(!done) {
		LogEvent("Printing plug-in in embed mode");

		ZeroMemory(&pd,sizeof(PRINTDLG));

		pd.lStructSize = sizeof(PRINTDLG);
		pd.hwndOwner = hWnd;
		pd.Flags = PD_RETURNDC|PD_NOSELECTION|PD_NOPAGENUMS|PD_USEDEVMODECOPIESANDCOLLATE;

		if(!PrintDlg(&pd))
			return;    // canceled by user

		ZeroMemory(&di,sizeof(DOCINFO));
		di.cbSize=sizeof(DOCINFO);
		di.lpszDocName= PRG " printout";
		di.lpszOutput=NULL;
		di.lpszDatatype=NULL;
		di.fwType=0;

		StartDoc(pd.hDC,&di);
		StartPage(pd.hDC);

		/* Plugins using a newer API expect an extra member at the end of NPWindow.
		 * This causes a problem with printing, because the NPPrint.print.embedPrint
		 * structure contains an NPWindow structure followed by the device context.
		 * Therefore different plugins expect the device context to be at different
		 * offsets. To workaround this I have defined two different NPPrint-like
		 * structures for the different APIs. API 0.9 uses the old style, 0.11 the
		 * new. (If there ever was an 0.10, I may have got it wrong.)
		 *
		 * However, if the user sets our version number to 0.9 or less, we'll always
		 * use the old style, to mimic an older browser.
		 */

		if(pl->Funcs.version <= 0x0009 || (pl->versionmajor==0 && pl->versionminor<=9)) {
			ZeroMemory(&nppr1,sizeof(NPPrint1));
			nppr1.mode=NP_EMBED;
			nppr1.print.embedPrint.window.window=NULL;
			nppr1.print.embedPrint.window.x=0;
			nppr1.print.embedPrint.window.y=0;
			nppr1.print.embedPrint.window.width=  GetDeviceCaps(pd.hDC,HORZRES);
			nppr1.print.embedPrint.window.height= GetDeviceCaps(pd.hDC,VERTRES);
			nppr1.print.embedPrint.platformPrint= (void*)(pd.hDC);
			pl->Funcs.print(pinst,(NPPrint*)(&nppr1));
		}
		else {
			ZeroMemory(&nppr2,sizeof(NPPrint2));
			nppr2.mode=NP_EMBED;
			nppr2.print.embedPrint.window.window=NULL;
			nppr2.print.embedPrint.window.x=0;
			nppr2.print.embedPrint.window.y=0;
			nppr2.print.embedPrint.window.width=  GetDeviceCaps(pd.hDC,HORZRES);
			nppr2.print.embedPrint.window.height= GetDeviceCaps(pd.hDC,VERTRES);
			nppr2.print.embedPrint.platformPrint= (void*)(pd.hDC);
			pl->Funcs.print(pinst,(NPPrint*)(&nppr2));
		}

		EndPage(pd.hDC);
		EndDoc(pd.hDC);

		DeleteDC(pd.hDC);
		if(pd.hDevMode)  GlobalFree(pd.hDevMode);
		if(pd.hDevNames) GlobalFree(pd.hDevNames);
	}

	SetCursor(hCur);
}

// Try to confuse the plugin by destroying its window and giving it another one.
void ReplaceWindow(void)
{
	HWND oldhWnd;

	if(pluginstate!=ST_RUNNING) return;
	if(!pl->iswindowed) {
		DisplayMsg(MSG_E,"Plug-in is windowless; no window to replace");
		return;
	}

	pl->window.window=NULL;
	MySetWindow();

	oldhWnd=pl->hWnd;

	// create the new window first, to make sure it has a different handle
	pl->hWnd = CreateWindow(
		"SOCKET45PLUGIN","a Plugin Window",WS_CHILD|WS_OVERLAPPED,
		0,0,50,50,hWndParent,NULL,hInst,NULL);

	if(!pl->hWnd) {
		DisplayMsg(MSG_S,"Can't create window for plug-in");
		return;
	}

	LogEvent("Created new plug-in window %04x",(int)(pl->hWnd));

	LogEvent("Destroying old plug-in window %04x",(int)(oldhWnd));
	DestroyWindow(oldhWnd);

	size_window();
	ShowWindow(pl->hWnd, SW_SHOW);

	send_setwindow();
}


// Send a(nother) stream to an existing plug-in. Pretty useless.
void StreamToRunningPlugin(HWND hWnd)
{
	char fn[MAX_PATH];

	if(pluginstate != ST_RUNNING) return;

	strcpy(fn,"");
	if(GetDatafileName(hWnd,fn)) {
		strcpy(pl->datafile,fn);
		WriteStream();
	}
}

int ReadSettings(HWND hWnd,char *fn, int bytype)
{
	DWORD datasize,datatype;
	HKEY key;
	long rv;

	DWORD tmpd;
	char ext[256];
	char szKey[256];

	if(bytype) {
		if(!GetFileExtension(fn,ext)) return 0;
		sprintf(szKey,"SOFTWARE\\Generic\\Socket45\\types\\%s",ext);
	}
	else {
		strcpy(ext,BaseName(pl->datafile));
		if(!strlen(ext)) return 0;
		sprintf(szKey,"SOFTWARE\\Generic\\Socket45\\files\\%s",ext);
	}

	if(RegOpenKeyEx(
		HKEY_CURRENT_USER,
		szKey,
		0,
		(REGSAM)KEY_READ,
		(PHKEY)(&key)
		)!=ERROR_SUCCESS) {
		return 0;
	}

	datasize=MAX_PATH;
	rv=RegQueryValueEx(key,"plugin",NULL,&datatype,(LPBYTE)pl->dllname,&datasize);

	datasize=128;
	rv=RegQueryValueEx(key,"mimetype",NULL,&datatype,(LPBYTE)pl->mimetype,&datasize);

	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"autourl",NULL,&datatype,(LPBYTE)&(pl->autourl),&datasize);
	if(!pl->autourl) {
		datasize=MAX_URL;
		rv=RegQueryValueEx(key,"url",NULL,&datatype,(LPBYTE)pl->url,&datasize);
	}

	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"autoagent",NULL,&datatype,(LPBYTE)&(pl->autoagent),&datasize);

	datasize=128;
	rv=RegQueryValueEx(key,"agent",NULL,&datatype,(LPBYTE)pl->useragent,&datasize);

	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"mode",NULL,&datatype,(LPBYTE)&(tmpd),&datasize);
	if(rv==ERROR_SUCCESS) pl->mode=(int16)tmpd;

	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"showurlreq",NULL,&datatype,(LPBYTE)&(pl->showurlreq),&datasize);
	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"autosize",NULL,&datatype,(LPBYTE)&(pl->autosize),&datasize);
	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"hidden",NULL,&datatype,(LPBYTE)&(pl->hidden),&datasize);
	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"width",NULL,&datatype,(LPBYTE)&(pl->width),&datasize);
	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"height",NULL,&datatype,(LPBYTE)&(pl->height),&datasize);
	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"statusbar",NULL,&datatype,(LPBYTE)&(pl->statusbar),&datasize);
	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"versionmajor",NULL,&datatype,(LPBYTE)&(pl->versionmajor),&datasize);
	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"versionminor",NULL,&datatype,(LPBYTE)&(pl->versionminor),&datasize);
	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"pfunctblsize",NULL,&datatype,(LPBYTE)&(pl->pfunctblsize),&datasize);
	datasize=sizeof(DWORD);
	rv=RegQueryValueEx(key,"nfunctblsize",NULL,&datatype,(LPBYTE)&(pl->nfunctblsize),&datasize);

	datasize=MAX_PARAMS;
	rv=RegQueryValueEx(key,"params",NULL,&datatype,(LPBYTE)(pl->params),&datasize);

	RegCloseKey(key);

	return 1;
}

void OpenDatafile(HWND hWnd, char *fn, int launch, int keepplugin)
{
	int rv;
	char tmppl[MAX_PATH];
	char buf[1024];

	strcpy(tmppl,pl->dllname);

	FactorySettings(pl);
	strcpy(pl->datafile,fn);

	// This is pretty messy. I should probably restructure it.
	// Then again, "If it ain't broke..."
	if(!ReadSettings(hWnd,fn,0)) {      // by filename
		if(!ReadSettings(hWnd,fn,1)) {    // by file type
			if(keepplugin && strlen(tmppl)) {
				sprintf(buf,"Do you want to use the current plug-in:\r\n  %s\r\nto view this file:\r\n  %s ?",
					BaseName(tmppl),fn);
				rv=MessageBox(hWndMain,buf,"Unknown file type",
					MB_YESNO|MB_ICONQUESTION|MB_DEFBUTTON2);
				if(rv==IDYES) {
					strcpy(pl->dllname,tmppl);
					goto here;
				}

			}
			if(!DialogBox(hInst,"SETUPDLG",hWnd,DlgProcSetup)) {  // otherwise ask the user
				return;
			}
		}
	}
here:
	if(launch)
		WriteStream();
}

void SelectAndOpenFile(HWND hWnd,int launch)
{
	char fn[MAX_PATH];

	strcpy(fn,"");
	if(!GetDatafileName(hWnd,fn)) {
		return;
	}

	UnloadPlugin();
	OpenDatafile(hWnd,fn,launch,1);
}


int InitProgram(void)
{
	// initialize our ugly globals...
	pl=(plugindata_t*)malloc(sizeof(plugindata_t));
	pinst=(NPP)malloc(sizeof(NPP_t));

	if(!pl || !pinst) {
		DisplayMsg(MSG_S,"Out of memory");
		return (-1);
	}

	FactorySettings(pl);
	strcpy(current_dllname,"");

	GetCurrentDirectory(MAX_PATH,datafile_dir);

	pluginstate=ST_OFF;

	return 0;
}

void ShutdownProgram(void)
{
	if(pluginstate != ST_OFF) UnloadPlugin();
	if(pl) free(pl);
	if(pinst) free(pinst);
}

// *********************************************************************
// *********

LRESULT CALLBACK WndProcParent(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	NPEvent ev;
	int tmp;

	id=LOWORD(wParam);

	switch(msg) {
	case WM_HSCROLL:
		switch(id) {
		case SB_LINELEFT: scrollpos.x-=20; break;
		case SB_LINERIGHT: scrollpos.x+=20; break;
		case SB_PAGELEFT: scrollpos.x-=80; break;
		case SB_PAGERIGHT: scrollpos.x+=80; break;
		case SB_THUMBPOSITION: scrollpos.x=HIWORD(wParam); break;
		case SB_THUMBTRACK: scrollpos.x=HIWORD(wParam); break;
		}
		set_scrollbars(1);
		return 0;
	case WM_VSCROLL:
		switch(id) {
		case SB_LINELEFT: scrollpos.y-=20; break;
		case SB_LINERIGHT: scrollpos.y+=20; break;
		case SB_PAGELEFT: scrollpos.y-=80; break;
		case SB_PAGERIGHT: scrollpos.y+=80; break;
		case SB_THUMBPOSITION: scrollpos.y=HIWORD(wParam); break;
		case SB_THUMBTRACK: scrollpos.y=HIWORD(wParam); break;
		}
		set_scrollbars(1);
		return 0;


	case WM_PAINT:
		{
			PAINTSTRUCT paintStruct;
			HDC hdc;
			RECT r;
			HPEN pen;
			int i,h,step;


			hdc = BeginPaint( hWnd, &paintStruct );
			if(pl->mode == NP_EMBED || !pl->iswindowed) {

				GetClientRect(hWnd,&r);

				pen=SelectObject(hdc,GetStockObject(BLACK_PEN));

				step=8;
				h=r.bottom-r.top;
				for(i=(r.left-h)-(r.left-h)%step;i<r.right;i+=step) {
					MoveToEx(hdc,i,0,NULL);
					LineTo(hdc,i+h,h);
				}

				SelectObject(hdc,pen);


				if(!pl->iswindowed) {
					ev.event=(uint16)msg;
					ev.wParam=wParam;
					ev.lParam=(LPARAM)hdc;
					pl->Funcs.event(pinst,&ev);
				}
			}
			EndPaint(hWnd, &paintStruct);

			return 0;
		}
	}


	if(!pl->iswindowed) {         // windowless

		switch(msg) {
		// pass on these messages to windowless plugins only
		// (WM_PAINT is handled above)
		case WM_LBUTTONDOWN: case WM_LBUTTONUP: case WM_LBUTTONDBLCLK:
		case WM_RBUTTONDOWN: case WM_RBUTTONUP: case WM_RBUTTONDBLCLK:
		case WM_MBUTTONDOWN: case WM_MBUTTONUP: case WM_MBUTTONDBLCLK:
		case WM_MOUSEMOVE:   case WM_KEYUP:     case WM_KEYDOWN:
		case WM_SETCURSOR:   case WM_SETFOCUS:  case WM_KILLFOCUS:

			ev.event=(uint16)msg;
			ev.wParam=wParam;
			ev.lParam=lParam;
			tmp=pl->Funcs.event(pinst,&ev);
			if(tmp) return 0;
			break;

		}
	}

	return (DefWindowProc(hWnd, msg, wParam, lParam));
}

// Does the file name start with NP and end with .DLL? 
int IsPlugin(char *fn)
{
	char tmp[32];
	char *b;

	GetFileExtension(fn,tmp);
	if(!_stricmp(tmp,"dll")) {
		b=BaseName(fn);
		if(strlen(b)<5) return 0;
		if( (b[0]=='n' || b[0]=='N') && 
			(b[1]=='p' || b[1]=='P') ) return 1;
	}
	return 0;
}

// Handle files dropped onto our program.
void DroppedFiles(HDROP hDrop)
{
	UINT rv;
	char fn[MAX_PATH];

	// get number of files dropped
	rv=DragQueryFile(hDrop,0xFFFFFFFF,NULL,0);
 
	if(rv==1) {
		rv=DragQueryFile(hDrop,0,fn,MAX_PATH);
		if(rv) {
			UnloadPlugin();

			if(IsPlugin(fn)) {
				FactorySettings(pl);
				strcpy(pl->dllname,fn);
			}
			else {
				OpenDatafile(hWndMain,fn,1,1);
			}
		}
			
	}
	else {
		DisplayMsg(MSG_E,"Cannot accept multiple dropped files");
	}

	DragFinish(hDrop);
}


LRESULT CALLBACK WndProcMain(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	UINT x,y,z;

	id=LOWORD(wParam);

	switch(msg) {
	case WM_CREATE:
		return InitProgram();

	case WM_DESTROY:
		ShutdownProgram();
		PostQuitMessage(0);
		return 0;

	case WM_ERASEBKGND:
		if(pluginstate==ST_RUNNING) return 1;
		break;

	case WM_PAINT: {
			PAINTSTRUCT paintStruct;
			HDC hdc;
			HGDIOBJ oldpen;
			RECT r;
			int i,h;
			const int step=16;

			hdc = BeginPaint( hWnd, &paintStruct );

			if(pluginstate!=ST_RUNNING) {
				GetClientRect(hWnd,&r);
				// draw cross-hatches
				oldpen=SelectObject(hdc,GetStockObject(BLACK_PEN));
				h=r.bottom-r.top;
				for(i=(r.left-h)-(r.left-h)%step;i<r.right;i+=step) {
					MoveToEx(hdc,i,0,NULL);
					LineTo(hdc,i+h,h);
				}
				for(i=(r.right+h)-(r.right+h)%step;i>0;i-=step) {
					MoveToEx(hdc,i,0,NULL);
					LineTo(hdc,i-h,h);
				}

				SelectObject(hdc,oldpen);
			}
			EndPaint( hWnd, &paintStruct );
			return 0;
		}

	case WM_DROPFILES:
		DroppedFiles((HDROP)wParam);
		return 0;

	case WM_SIZE:
		if(pluginstate == ST_RUNNING) {
			size_window();
			if(pl->mode!=NP_EMBED || pl->autosize) {
				send_setwindow();     // maybe not?
			}
		}
		return 0;

	case WM_INITMENU:
		// Figure out which menu items need to be disabled.
		// This seems overly complicated...

		x= (strlen(pl->datafile)>0) ?1:0;
		y= (strlen(pl->dllname)>0) ?1:0;
		z= ( (x==MF_ENABLED) && (pluginstate!=ST_OFF || (y==MF_ENABLED)) ) ?1:0;
		
		EnableMenuItem((HMENU)wParam,ID_START,MF_BYCOMMAND| 
			(y?MF_ENABLED:MF_GRAYED));
		EnableMenuItem((HMENU)wParam,ID_STOP,MF_BYCOMMAND| 
			(pluginstate!=ST_OFF)?MF_ENABLED:MF_GRAYED);

		EnableMenuItem((HMENU)wParam,ID_INITIALIZE,MF_BYCOMMAND| 
			((pluginstate==ST_OFF && y)?MF_ENABLED:MF_GRAYED));

		EnableMenuItem((HMENU)wParam,ID_NEW,MF_BYCOMMAND|
			(pluginstate!=ST_RUNNING && y)?MF_ENABLED:MF_GRAYED);

		EnableMenuItem((HMENU)wParam,ID_WRITE,MF_BYCOMMAND|
			(x && (pluginstate!=ST_OFF || (pluginstate==ST_OFF && y)))?MF_ENABLED:MF_GRAYED);

		EnableMenuItem((HMENU)wParam,ID_DESTROY,MF_BYCOMMAND|(pluginstate==ST_RUNNING)?MF_ENABLED:MF_GRAYED);
		EnableMenuItem((HMENU)wParam,ID_SHUTDOWN,MF_BYCOMMAND|(pluginstate!=ST_OFF)?MF_ENABLED:MF_GRAYED);

		EnableMenuItem((HMENU)wParam,ID_REPLACEWINDOW,MF_BYCOMMAND|(pluginstate==ST_RUNNING)?MF_ENABLED:MF_GRAYED);

		EnableMenuItem((HMENU)wParam,ID_NEWSTREAM,MF_BYCOMMAND|(pluginstate==ST_RUNNING)?MF_ENABLED:MF_GRAYED);
		EnableMenuItem((HMENU)wParam,ID_PRINT,MF_BYCOMMAND|(pluginstate==ST_RUNNING)?MF_ENABLED:MF_GRAYED);

		EnableMenuItem((HMENU)wParam,ID_SAVEBYTYPE,MF_BYCOMMAND|x?MF_ENABLED:MF_GRAYED);
		EnableMenuItem((HMENU)wParam,ID_SAVEBYFILE,MF_BYCOMMAND|x?MF_ENABLED:MF_GRAYED);
		EnableMenuItem((HMENU)wParam,ID_ERASESETTINGS,MF_BYCOMMAND|x?MF_ENABLED:MF_GRAYED);
		
		CheckMenuItem((HMENU)wParam,ID_STBAR,MF_BYCOMMAND|(pluginstate==ST_RUNNING && pl->statusbar)?MF_CHECKED:MF_UNCHECKED);
		EnableMenuItem((HMENU)wParam,ID_STBAR,MF_BYCOMMAND|(pluginstate==ST_RUNNING)?MF_ENABLED:MF_GRAYED);

		CheckMenuItem((HMENU)wParam,ID_LOG,MF_BYCOMMAND|(logging)?MF_CHECKED:MF_UNCHECKED);
		EnableMenuItem((HMENU)wParam,ID_CLEARLOG,MF_BYCOMMAND|(logging)?MF_ENABLED:MF_GRAYED);
		return 0;

	case WM_USER:
		if(strlen(filetoopen)) {
			OpenDatafile(hWnd,filetoopen,1,0);
		}
		return 0;

	case WM_COMMAND:
		switch(LOWORD(wParam)) {
		case ID_EXIT:
			DestroyWindow(hWnd);
			return 0;
		case ID_ABOUT:
			DialogBox(hInst,"ABOUTDLG",hWnd,DlgProcAbout);
			return 0;
		case ID_HOMEPAGE:
			ShellExecute(hWndMain,"open", HOMEPAGE, "", NULL, SW_SHOWNORMAL);
			return 0;
		case ID_CLEARLOG:
			if(logging)
				SendMessage(hWndList,LB_RESETCONTENT,0,0);
			return 0;
		case ID_LOG:
			if(logging) { // close log window
				if(IsWindow(hWndLog)) {
					DestroyWindow(hWndLog);
				}
				logging=0;
			}
			else { // open log window
				RECT rw;

				SystemParametersInfo(SPI_GETWORKAREA,(UINT)0,(PVOID)&rw,(UINT)0);

				hWndLog=CreateWindow(
					"SOCKET45LOG","Socket45 Event Log",
					WS_VISIBLE|WS_OVERLAPPEDWINDOW,
					rw.left,                        // horizontal position
					rw.bottom-(rw.bottom-rw.top)/4,	// vertical position
					(2*(rw.right-rw.left))/3,	    // width
					(rw.bottom-rw.top)/4,           // height
					hWnd,NULL,hInst,NULL);
			}
			return 0;

		case ID_OPEN:
			SelectAndOpenFile(hWnd,1);
			return 0;
		case ID_LOAD:
			SelectAndOpenFile(hWnd,0);
			return 0;
		case ID_NEWSTREAM:
			StreamToRunningPlugin(hWnd);
			return 0;
		case ID_PRINT:
			PrintPlugin(hWnd);
			return 0;

		case ID_SETUP:
			if(DialogBox(hInst,"SETUPDLG",hWnd,DlgProcSetup) == 2) {
				// user chose Start, so (re)start the plug-in
				UnloadPlugin();
				WriteStream();
			}
			return 0;

		case ID_INITIALIZE:
			LoadPlugin();
			return 0;

		case ID_NEW:
			NewPluginInstance();
			return 0;

		case ID_WRITE:
			WriteStream();
			return 0;

		case ID_REPLACEWINDOW:
			ReplaceWindow();
			return 0;

		case ID_DESTROY:
			DestroyPluginInstance();
			return 0;

		case ID_SHUTDOWN:
			UnloadPlugin();
			return 0;

		case ID_START:
			DestroyPluginInstance();
			WriteStream();
			return 0;

		case ID_STOP:
			UnloadPlugin();
			return 0;

		case ID_SAVEBYTYPE:
			SaveSettings(hWnd,1);
			return 0;
		case ID_SAVEBYFILE:
			SaveSettings(hWnd,0);
			return 0;
		case ID_ERASESETTINGS:
			EraseSettings(hWnd);
			return 0;
		case ID_STBAR:
			pl->statusbar = !pl->statusbar;
			if(pluginstate == ST_RUNNING) {
				size_window();
				send_setwindow();
			}
			return 0;
		}

	}

	return (DefWindowProc(hWnd, msg, wParam, lParam));
}

// This function will be overridden (subclassed) by the plug-in,
// so it doesn't really need to do much of anything.
LRESULT CALLBACK WndProcPlugin(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	if(pl->iswindowed) {
		switch(msg) {
		case WM_PAINT: 
			{
				PAINTSTRUCT paintStruct;
				HDC hdc;

				hdc = BeginPaint( hWnd, &paintStruct );
				EndPaint( hWnd, &paintStruct );
				return 0;
			}
		}
	}
	return (DefWindowProc(hWnd, msg, wParam, lParam));
}

// Enable or disable dialog controls based on the contents of other dialog
// controls.
void EnableDisableStuff(HWND hWnd)
{
	int embedded,fullpage,autourl,autoagent,autosize,hidden;
	
	embedded= (IsDlgButtonChecked(hWnd,IDC_EMBEDDED)==BST_CHECKED);
	fullpage= (IsDlgButtonChecked(hWnd,IDC_FULLPAGE)==BST_CHECKED);
	autourl = (IsDlgButtonChecked(hWnd,IDC_AUTOURL )==BST_CHECKED);
	autosize= (IsDlgButtonChecked(hWnd,IDC_AUTOSIZE)==BST_CHECKED);
	hidden=   (IsDlgButtonChecked(hWnd,IDC_HIDDEN)==BST_CHECKED);
	autoagent=(IsDlgButtonChecked(hWnd,IDC_AUTOAGENT)==BST_CHECKED);

	EnableWindow(GetDlgItem(hWnd,IDC_URL),!autourl);
	EnableWindow(GetDlgItem(hWnd,IDC_USERAGENT),!autoagent);
	EnableWindow(GetDlgItem(hWnd,IDC_PARAMS),!fullpage);
	EnableWindow(GetDlgItem(hWnd,IDC_HIDDEN),!fullpage);
	EnableWindow(GetDlgItem(hWnd,IDC_AUTOSIZE),!fullpage && !hidden);
	EnableWindow(GetDlgItem(hWnd,IDC_WIDTH),!autosize && !hidden && !fullpage);
	EnableWindow(GetDlgItem(hWnd,IDC_HEIGHT),!autosize && !hidden && !fullpage);
}

// Read the Version info from the plugin to determine what MIME types it handles,
// and load them into the given combo box.
void SetMIMETypeCombo(HWND hWnd,int control, char *fn)
{
	LONG rv;
	int i;
	DWORD h;
	DWORD size;
	char *vbuf;
	char mbuf[1024];
	char *ibuf;
	char *t;
	UINT len;
	BOOL r;

	// Clear the listbox. (can't use CB_RESETCONTENT, it clears the edit control too)
	rv=SendDlgItemMessage(hWnd,control,CB_GETCOUNT,0,0);
	if(rv==CB_ERR) return;
	for(i=0;i<rv;i++) {
		SendDlgItemMessage(hWnd,control,CB_DELETESTRING,0,0);
	}

	if(!strlen(fn)) return;

	size=GetFileVersionInfoSize(fn, &h);
	if(!size) return;

	vbuf=malloc(size);
	if(!vbuf) return;

	if(GetFileVersionInfo(fn,h,size,vbuf)) {
		r=VerQueryValue(vbuf,TEXT("\\StringFileInfo\\040904E4\\MIMEType"),
			&ibuf,&len);
		if(r && len) {
			strncpy(mbuf,ibuf,1024); mbuf[1023]='\0';
			t=strtok(mbuf,"|");
			while(t) {
				SendDlgItemMessage(hWnd,control,CB_ADDSTRING,0,(LPARAM)t);
				t=strtok(NULL,"|");
			}
			// SendDlgItemMessage(hWnd,control,CB_SETCURSEL,0,0);
		}

	}
	free(vbuf);
}

BOOL CALLBACK DlgProcSetup(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	WORD id;
	char fn[MAX_PATH];

	id=LOWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:
		SetDlgItemText(hWnd,IDC_PLUGIN,pl->dllname);
		SetDlgItemText(hWnd,IDC_DATAFILE,pl->datafile);
		SetDlgItemText(hWnd,IDC_MIMETYPE,pl->mimetype);
		SetDlgItemText(hWnd,IDC_URL,pl->url);
		CheckDlgButton(hWnd,IDC_EMBEDDED,pl->mode==NP_EMBED?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hWnd,IDC_FULLPAGE,pl->mode==NP_FULL?BST_CHECKED:BST_UNCHECKED);
		SetDlgItemText(hWnd,IDC_PARAMS,pl->params);

		CheckDlgButton(hWnd,IDC_SHOWURLREQ,pl->showurlreq?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hWnd,IDC_AUTOURL,pl->autourl?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hWnd,IDC_AUTOAGENT,pl->autoagent?BST_CHECKED:BST_UNCHECKED);

		CheckDlgButton(hWnd,IDC_AUTOSIZE,pl->autosize?BST_CHECKED:BST_UNCHECKED);
		CheckDlgButton(hWnd,IDC_HIDDEN,pl->hidden?BST_CHECKED:BST_UNCHECKED);
		SetDlgItemInt(hWnd,IDC_WIDTH,pl->width,FALSE);
		SetDlgItemInt(hWnd,IDC_HEIGHT,pl->height,FALSE);
		SetDlgItemInt(hWnd,IDC_VERSIONMAJ,pl->versionmajor,FALSE);
		SetDlgItemInt(hWnd,IDC_VERSIONMIN,pl->versionminor,FALSE);
		SetDlgItemInt(hWnd,IDC_PFUNCTBLSIZE,pl->pfunctblsize,FALSE);
		SetDlgItemInt(hWnd,IDC_NFUNCTBLSIZE,pl->nfunctblsize,FALSE);

		EnableDisableStuff(hWnd);
		SetMIMETypeCombo(hWnd,IDC_MIMETYPE,pl->dllname);

		SendDlgItemMessage(hWnd,IDC_USERAGENT,CB_ADDSTRING,0,(LPARAM)USERAGENT);
		SendDlgItemMessage(hWnd,IDC_USERAGENT,CB_ADDSTRING,0,(LPARAM)"Mozilla/4.5 [en] (Win98; I)");
		SendDlgItemMessage(hWnd,IDC_USERAGENT,CB_ADDSTRING,0,(LPARAM)"Mozilla/3.04 [en] (Win95; I)");
		SendDlgItemMessage(hWnd,IDC_USERAGENT,CB_ADDSTRING,0,(LPARAM)"Microsoft Internet Explorer");
		SetDlgItemText(hWnd,IDC_USERAGENT,pl->useragent);
		return 1;

	case WM_COMMAND:
		switch(id) {
		case IDC_EMBEDDED:
		case IDC_FULLPAGE:
		case IDC_AUTOURL:
		case IDC_AUTOAGENT:
		case IDC_AUTOSIZE:
		case IDC_HIDDEN:
			if(HIWORD(wParam)==BN_CLICKED)
				EnableDisableStuff(hWnd);
			break;

		case IDC_BROWSE_DATA:
			GetDlgItemText(hWnd,IDC_DATAFILE,fn,MAX_PATH);
			if(GetDatafileName(hWnd,fn)) {
				SetDlgItemText(hWnd,IDC_DATAFILE,fn);
			}
			break;

		case IDC_BROWSE_PLUGIN:
			GetDlgItemText(hWnd,IDC_PLUGIN,fn,MAX_PATH);
			if(GetPluginName(hWnd,fn)) {
				SetDlgItemText(hWnd,IDC_PLUGIN,fn);
				SetMIMETypeCombo(hWnd,IDC_MIMETYPE,fn);
			}
			break;

		case IDOK:
		case IDC_APPLY:
		case IDC_START:
			GetDlgItemText(hWnd,IDC_PLUGIN,pl->dllname,MAX_PATH);
			GetDlgItemText(hWnd,IDC_DATAFILE,pl->datafile,MAX_PATH);
			GetDlgItemText(hWnd,IDC_MIMETYPE,pl->mimetype,128);
			GetDlgItemText(hWnd,IDC_URL,pl->url,MAX_URL);
			GetDlgItemText(hWnd,IDC_USERAGENT,pl->useragent,128);
			if(IsDlgButtonChecked(hWnd,IDC_FULLPAGE)==BST_CHECKED) pl->mode=NP_FULL;
			if(IsDlgButtonChecked(hWnd,IDC_EMBEDDED)==BST_CHECKED) pl->mode=NP_EMBED;
			GetDlgItemText(hWnd,IDC_PARAMS,pl->params,MAX_PARAMS);

			pl->showurlreq=(IsDlgButtonChecked(hWnd,IDC_SHOWURLREQ)==BST_CHECKED);
			pl->autourl=(IsDlgButtonChecked(hWnd,IDC_AUTOURL)==BST_CHECKED);
			pl->autoagent=(IsDlgButtonChecked(hWnd,IDC_AUTOAGENT)==BST_CHECKED);

			pl->autosize=(IsDlgButtonChecked(hWnd,IDC_AUTOSIZE)==BST_CHECKED);
			pl->hidden=(IsDlgButtonChecked(hWnd,IDC_HIDDEN)==BST_CHECKED);
			pl->width=GetDlgItemInt(hWnd,IDC_WIDTH,NULL,FALSE);
			pl->height=GetDlgItemInt(hWnd,IDC_HEIGHT,NULL,FALSE);
			pl->versionmajor=GetDlgItemInt(hWnd,IDC_VERSIONMAJ,NULL,FALSE);
			pl->versionminor=GetDlgItemInt(hWnd,IDC_VERSIONMIN,NULL,FALSE);
			pl->pfunctblsize=GetDlgItemInt(hWnd,IDC_PFUNCTBLSIZE,NULL,FALSE);
			pl->nfunctblsize=GetDlgItemInt(hWnd,IDC_NFUNCTBLSIZE,NULL,FALSE);

			if(id==IDC_APPLY) {
				UnloadPlugin();
				WriteStream();
			}

			if(id==IDC_START) {
				EndDialog(hWnd, 2);
			}
			if(id==IDOK) {
				EndDialog(hWnd, 1);
			}
			return 1;

		case IDCANCEL:
			EndDialog(hWnd, 0);
			return 1;
		}
	}
	return 0;
}


BOOL CALLBACK DlgProcAbout(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	char buf[1024];
	WORD id;
	id=LOWORD(wParam);

	switch (msg) {
	case WM_INITDIALOG:
		sprintf(buf,"%s: Browser plug-in container\r\nVersion %s -- %s\r\n\r\nBy Jason Summers\r\n%s"
			,PRG,PRGVERSION,__DATE__,HOMEPAGE);
		SetDlgItemText(hWnd,IDC_ABOUTTEXT,buf);
		return 1;
	case WM_COMMAND:
		switch(id) {
		case IDOK:
		case IDCANCEL:
			EndDialog(hWnd, 0);
			return 1;
		}
	}
	return 0;		// Didn't process a message
}




LRESULT CALLBACK WndProcLog(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	RECT r;

	switch(msg) {
	case WM_CREATE:
		GetClientRect(hWnd,&r);

		// create the listbox that covers the whole client area
		hWndList=CreateWindow(
			"LISTBOX",
			"Log",
			// Might as well set NOINTEGRALHEIGHT; Windows sizes it incorrectly anyway when
			// there's a horiz. scrollbar.
			WS_VISIBLE|WS_CHILD|WS_VSCROLL|WS_HSCROLL|LBS_HASSTRINGS|LBS_NOINTEGRALHEIGHT,
			r.left, r.top, r.right-r.left, r.bottom-r.top,
			hWnd,NULL,hInst,NULL);

		SendMessage(hWndList,WM_SETFONT,(WPARAM)GetStockObject(ANSI_VAR_FONT),MAKELPARAM(TRUE,0));
		SendMessage(hWndList,LB_SETHORIZONTALEXTENT,1500,0);

		logging=1;
		return 0;

	case WM_SIZE:
		GetClientRect(hWnd,&r);
		SetWindowPos(hWndList,NULL,r.left,r.top,
			r.right-r.left,r.bottom-r.top,SWP_NOZORDER);
		return 0;

	case WM_DESTROY:
		logging=0;
		return 0;
	
	}
	return (DefWindowProc(hWnd, msg, wParam, lParam));
}


int WINAPI WinMain(HANDLE hInstance, HANDLE hPrevInstance, LPSTR lpCmdLine,
						  int nCmdShow)
{
	MSG msg;
	char buf[256];
	RECT r;

	HACCEL hAccTable;
	InitCommonControls();

	hInst = hInstance;

	// Read initial datafile from command line and strip of quotes if it has them.
	// This needs work. It is q&d.
	strcpy(filetoopen,"");
	strncpy(buf,lpCmdLine,255); buf[255]='\0';
	if(buf[0]=='"') {
		strcpy(filetoopen,&buf[1]);
		filetoopen[strlen(filetoopen)-1]='\0';
	}
	else {
		strcpy(filetoopen,buf);
	}

	hAccTable=LoadAccelerators(hInst,"ACCELTABLE");
	RegisterClasses();

	hWndMain = CreateWindow(
		"SOCKET45MAIN",PRG,
		WS_VISIBLE|WS_OVERLAPPEDWINDOW|WS_CLIPCHILDREN,
		CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,
		NULL,NULL,hInst,NULL);
	if (!hWndMain) return 0;

	hWndStBar=CreateStatusWindow(WS_CHILD,"",hWndMain,69);

	GetWindowRect(hWndStBar,&r);
	sbheight=r.bottom-r.top;

	DragAcceptFiles(hWndMain,TRUE);

	PostMessage(hWndMain,WM_USER,0,0);

	while(GetMessage(&msg,NULL,0,0)){
		if (!TranslateAccelerator(hWndMain, hAccTable, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}
	return (msg.wParam);
}



//==============================================================
// Functions the plugin can call. Many of them don't do anything.
// Maybe I should have named them to start with NPN_, but the
// point is that it doesn't make any difference.

static char CALLFROM[] = "Call from plug-in";
static char NOTIMP[] = "[Not implemented]";

NPError S45_GetURL(NPP instance, const char* url, const char* window)
{
	char msg[1024];
	int rv;

	LogEvent("%s: GetURL (URL:%s, TARGET:%s) %s",CALLFROM,url,window?window:"",NOTIMP);

	if(pl->showurlreq) {
		if(AbsoluteURL(url)) {
			sprintf(msg,"GetURL\r\nURL: %s\r\nTARGET: %s\r\n\r\nDo you want to open this page in your browser?",url,window?window:"");
			rv=MessageBox(hWndMain,msg,CALLFROM,MB_YESNO|MB_ICONQUESTION|MB_DEFBUTTON2);
			if(rv==IDYES) {
				ShellExecute(hWndMain,"open", url, "", NULL, SW_SHOWNORMAL);
			}
		}
		else {
			sprintf(msg,"GetURL\r\nURL: %s\r\nTARGET: %s",url,window?window:"");
			rv=MessageBox(hWndMain,msg,CALLFROM,MB_OK|MB_ICONINFORMATION);
		}
	}

	//return NPERR_GENERIC_ERROR;
	return NPERR_NO_ERROR;  // pretend it succeeded
}

NPError S45_PostURL(NPP instance, const char* url, const char* window, uint32 len, const char* buf, NPBool file)
{	
	char msg[1024];
	LogEvent("%s: PostURL (URL:%s, TARGET:%s) %s",CALLFROM,url,window?window:"",NOTIMP);
	if(pl->showurlreq) {
		sprintf(msg,"PostURL\r\nURL: %s\r\nTARGET: %s",url,window?window:"");
		MessageBox(hWndMain,msg,CALLFROM,MB_OK|MB_ICONINFORMATION);
	}
	return NPERR_GENERIC_ERROR;
}

NPError S45_RequestRead(NPStream* stream, NPByteRange* rangeList)
{	
	LogEvent("%s: RequestRead %s",CALLFROM,NOTIMP);
	return NPERR_GENERIC_ERROR;
}

NPError S45_NewStream(NPP instance, NPMIMEType type, const char* window, NPStream** stream)
{	
	LogEvent("%s: NewStream %s",CALLFROM,NOTIMP);
	return NPERR_GENERIC_ERROR;
}

int32 S45_Write(NPP instance, NPStream* stream, int32 len, void* buffer)
{	
	LogEvent("%s: Write %s",CALLFROM,NOTIMP);
	return (int32)0;
}

NPError S45_DestroyStream(NPP instance, NPStream* stream, NPReason reason)
{	
	LogEvent("%s: DestroyStream %s",CALLFROM,NOTIMP);
	return NPERR_GENERIC_ERROR;
}

void    S45_Status(NPP instance, const char* message)
{
//	if(pl->statusbar)
	if(hWndStBar)
		SetWindowText(hWndStBar,message?message:"");

	if(!pl->statusbar)
		LogEvent("Status: %s",message?message:"(null)");
	return;
}

const char*	 S45_UserAgent(NPP instance)
{
	char *u;

	if(pl->autoagent)
		u=USERAGENT;
	else
		u=pl->useragent;

	LogEvent("Plug-in asked for UserAgent; telling it \"%s\"",u);

	return u;
}

void*   S45_MemAlloc(uint32 size)
{
	void *m;
	m=malloc(size);

	LogEvent("Plug-in allocated %d bytes of memory at %p",size,m);
	return m;
}

void    S45_MemFree(void* ptr)
{
	LogEvent("Plug-in freed memory at %p",ptr);
	free(ptr);
}

uint32  S45_MemFlush(uint32 size)
{
	LogEvent("%s: MemFlush",CALLFROM);
	return (uint32)0;
}

void    S45_ReloadPlugins(NPBool reloadPages)
{
	LogEvent("%s: ReloadPlugins [this shouldn't happen]",CALLFROM);
	return;
}

JRIEnv* S45_GetJavaEnv(void)
{
	LogEvent("%s: GetJavaEnv %s",CALLFROM,NOTIMP);
	return NULL;
}

jref    S45_GetJavaPeer(NPP instance)
{
	LogEvent("%s: GetJavaPeer %s",CALLFROM,NOTIMP);
	return NULL;
}

NPError S45_GetURLNotify(NPP instance, const char* url, const char* window, void* notifyData)
{	
	char msg[1024];
	LogEvent("%s: GetUrlNotify (URL:%s, TARGET:%s) %s",CALLFROM,url,window?window:"",NOTIMP);
	if(pl->showurlreq) {
		sprintf(msg,"GetURLNotify\r\nURL: %s\r\nTARGET: %s",url,window?window:"");
		MessageBox(hWndMain,msg,CALLFROM,MB_OK|MB_ICONINFORMATION);
	}
	return NPERR_GENERIC_ERROR;
}

NPError S45_PostURLNotify(NPP instance, const char* url, const char* window, uint32 len, const char* buf, NPBool file, void* notifyData)
{	
	char msg[1024];
	
	LogEvent("%s: PostUrlNotify (URL:%s, TARGET:%s) %s",CALLFROM,url,window?window:"",NOTIMP);
	if(pl->showurlreq) {
		sprintf(msg,"PostURLNotify\r\nURL: %s\r\nTARGET: %s",url,window?window:"");
		MessageBox(hWndMain,msg,CALLFROM,MB_OK|MB_ICONINFORMATION);
	}
	return NPERR_GENERIC_ERROR;
}

NPError S45_GetValue(NPP instance, NPPVariable variable, void *ret_value)
{
	HWND *tmp;

	switch(variable) {
	case NPNVnetscapeWindow:   // the only value we recognize
		tmp=(HWND*)(ret_value);
		(*tmp)= hWndParent;

		LogEvent("Plug-in requested window handle");
		return NPERR_NO_ERROR;
	}

	LogEvent("%s: GetValue %s",CALLFROM,NOTIMP);
	return NPERR_GENERIC_ERROR;
}

NPError S45_SetValue(NPP instance, NPNVariable variable, void *value)
{
	int tmp;

	switch(variable) {
	case NPPVpluginWindowBool:
		if(pl->where_are_we==1) {
			tmp=(int)(value);
			pl->iswindowed= tmp?1:0;
			if(pl->iswindowed)
				LogEvent("Plug-in superfluously informed us that it is windowed");
			else
				LogEvent("Plug-in informed us that it is windowless");
			return NPERR_NO_ERROR;
		}
		else {
			LogEvent("Plug-in error: Call to SetValue(Windowless) outside of NPP_New");
		}
		break;
	case NPPVpluginTransparentBool:
		if(!pl->iswindowed) {
			tmp=(int)(value);
			pl->istransparent=tmp?1:0;
			if(pl->istransparent)
				LogEvent("Plug-in informed us that it is transparent");
			else
				LogEvent("Plug-in informed us that it is opaque");
			return NPERR_NO_ERROR;
		}
		else {
			LogEvent("Plug-in error: Call to SetValue(Transparency) for windowed plug-in");
		}
		break;
	default:
		LogEvent("%s: SetValue %d %s",CALLFROM,(int)variable,NOTIMP);
	}
	return NPERR_GENERIC_ERROR;
}

void S45_InvalidateRect(NPP instance, NPRect *rect)
{
	RECT r;

	// not sure if this is relative to the plug-ins position or its parent window's
	r.left  =rect->left;
	r.top   =rect->top;
	r.right =rect->right;
	r.bottom=rect->bottom;

	LogEvent("%s: InvalidateRect(L:%d, T:%d, R:%d, B:%d)",CALLFROM,
		rect->left, rect->top, rect->right, rect->bottom);
	InvalidateRect(hWndParent,&r,TRUE);
}

void S45_InvalidateRegion(NPP instance, NPRegion region)
{
	LogEvent("%s: InvalidateRegion",CALLFROM);
	InvalidateRgn(hWndParent,(HRGN)region,TRUE);
}

void S45_ForceRedraw(NPP instance)
{
	LogEvent("%s: ForceRedraw",CALLFROM);
	InvalidateRect(hWndParent,NULL,TRUE);
}
